/* 
    IScratch Project
    Published under GPL license
*/

#include <stdrom.h>
#include <define.h>
#include <libc.h>
#include <L_libc.h>

#include "icons.h"
#include "modehook.h"

/* Code of regions */
#define OBJ_IC_PAD       0x9000
#define OBJ_IC_MENU      0x9001
#define OBJ_IC_SCREEN    0x9002

#define OBJ_IC_MENU_UP   0x9003
#define OBJ_IC_MENU_DOWN 0x9004
#define OBJ_IC_MENU_L1   0x9005
#define OBJ_IC_MENU_L2   0x9006
#define OBJ_IC_MENU_L3   0x9007
#define OBJ_IC_MENU_L4   0x9008

#define OBJ_IC_PSAVE     0x9009
#define OBJ_IC_PCLOSE    0x900A
#define OBJ_IC_PEDIT     0x900B

/* Filename */
#define FILENAME "IDraw"

/* Used on painting screen*/
TCHTBL TchListDraw[] = {
  0,0,159,149,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_MOVE_OUT | ACT_BREAK,
  OBJ_IC_PAD,
  0x0000,

  0,150,159,159,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_MOVE_OUT,
  OBJ_IC_MENU,
  0x0000,

  0, 0, 0, 0,
  ACT_NONE,
  OBJ_END,
  0x0000
};

/* Full screen (without hardicons) */
TCHTBL TchListFull[] = {
  0,0,159,159,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_SCREEN,
  0x0000,

  0, 0, 0, 0,
  ACT_NONE,
  OBJ_END,
  0x0000
};

/* Record select screen */
TCHTBL TchListMenu[] = {
  0,0,159,15,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_MENU_UP,
  0x0000,

  0,16,159, 47,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK | ACT_MOVE_IN | ACT_MOVE_OUT,
  OBJ_IC_MENU_L1,
  0x0000,

  0,48,159,79,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK | ACT_MOVE_IN | ACT_MOVE_OUT,
  OBJ_IC_MENU_L2,
  0x0000,

  0,80,159,111,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK | ACT_MOVE_IN | ACT_MOVE_OUT,
  OBJ_IC_MENU_L3,
  0x0000,

  0,112,159,143,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK | ACT_MOVE_IN | ACT_MOVE_OUT,
  OBJ_IC_MENU_L4,
  0x0000,

  0,144,159,159,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_MENU_DOWN,
  0x0000,


  0, 0, 0, 0,
  ACT_NONE,
  OBJ_END,
  0x0000
};

/* Exit confirmation panel */
TCHTBL TchListPanel[] = {
  32,62,61,87,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_PSAVE,
  0x0000,

  64,62,93,87,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_PCLOSE,
  0x0000,

  96,62,125,87,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_PEDIT,
  0x0000,

  0,0,159,159,
  ACT_MOVE | ACT_MAKE | ACT_BREAK_IN | ACT_BREAK,
  OBJ_IC_SCREEN,
  0x0000,

  0, 0, 0, 0,
  ACT_NONE,
  OBJ_END,
  0x0000
};

/* Variable for storing event info */
TCHSTS g_tch;

/* Tools */
#define PEN_ERASE 0
#define PEN_DRAW 1
#define PEN_CORRECT 2

/* Styles of pen/eraser */
#define STYLE_CURVE    0
#define STYLE_LINE     1
#define STYLE_STRAIGHT 2
#define STYLE_BOX      3
#define STYLE_ELLIPSE  4

#define LAST_STYLE     STYLE_ELLIPSE

/* Status data */
typedef struct {
  word line;                /* Current line       2*/
  word lines;               /* Total lines        2*/
  word buf_size;            /* Buffer size        2*/
  byte pen_type;            /* Draw/erase/correct 1*/
  byte pen_mesh;            /* Solid/mesh         1*/
  byte pen_size;            /* Width of pen       1*/
  byte eraser_size;         /* Width of eraser    1*/
  /* end of ver1 header                   Total: 10 bytes*/
  byte ver2sign[6];         /* Sign of IS>=ver2   6*/
  byte draw_style;          /* Draw style         1*/
  byte erase_style;         /* Erase style        1*/
  /* end of ver2 header                   Total:  8 bytes*/     
  byte extension[120];
} t_stat; /* Total 138 bytes*/

#define MAX_BUFFER_LEN (30720-sizeof(t_stat))

/* Default header of new record*/
byte far default_header[]=
{0,0,0,0,0,0,PEN_DRAW,0,1,2,'I','S','v','2','.','0',STYLE_CURVE,STYLE_CURVE};

/* Internal screen buffer */
byte g_scr[3004];
/* Buffer for record title */
byte g_file_scr[4+30*20];
/* Pointers (used for fast scrolling) */
word points[480];
/* Last used pointer */
word level=0;
/* Buffer for QuickMemo exchange */
SHOT_INF g_shot;
/* QuickMemo flag */
byte call_shot=1;

/* Flags */
#define LINE_COMPRESSED 0xE0
#define LINE_FREE 0xC0
#define BUFFER_END 0xFF

/* Filebuffer */
LFILE_BUF far fb;
/* File descriptor */
FILE_INFO finf;
/* Pointer to image data */
byte far *const g_buf = fb.fbuf.bin.bin_buf + sizeof(t_stat);
/* Pointer to status data */
t_stat far *const g_stat = (t_stat far *const)fb.fbuf.bin.bin_buf;

/* Mode&submode of IDraw file */
byte mode;
byte submode;
/* Flag - if user select creating of new record */
byte new_file;
/* First visible record */
int file = 0;
/* Current selected record */
int sel = 0;
/* Flag - file was changed */
byte need_save=0;
/* Flag - file was changed while scrolling */
byte need_save_scroll=0;
/* Flag - we are in painting mode */
byte full_screen=1;
/* Flag - is current record in new format */
/*byte isVer2=1;*/ /* Reserved. Don't remove. */
/* Flag - is dialog on screen */
byte dialog_mode=0;

void set_touch(TCHTBL* tbl)
{/* Setting touch  information */
  LibTchStackClr();
  LibTchStackPush(NULL);
  LibTchStackPush(TchHardIcon);
  LibTchStackPush(TchActionKey);
  LibTchStackPush(tbl);
  LibTchInit();
}

void correct_settings(void)
{/* Checking status data of record */
  int i=10;
  byte isVer2=1;
  for(;i<16; i++)
    if (((byte far*)g_stat)[i]!=default_header[i])
      isVer2=0;
  if(!isVer2||g_stat->draw_style>LAST_STYLE)
  {
    g_stat->draw_style=STYLE_CURVE;
  }
  if(!isVer2||g_stat->erase_style>LAST_STYLE)
  {
    g_stat->erase_style=STYLE_CURVE;
  }
  g_stat->buf_size = fb.fbuf.bin.char_num-1-sizeof(t_stat);
  points[0]=0;
  level=0;
}


word memo_init(void) 
{/* Search IDraw-file*/
  byte *filename=FILENAME;
  if (!LibGetAllEntry(filename, &mode, &submode)) 
  {
    if (!LibSubEntrySave(filename, &submode)) 
      return 0;
    LibGetAllEntry(filename, &mode, &submode);
  }
  return 1;
}

void memo_reset(void)
{/* Reset file descriptor */
  fb.fsb_main_entry_ = mode;
  fb.fsb_sub_entry_ = submode;
  fb.fsb_scrt_info_ = 0x80;
  finf.fp = 0xFFFE;
  finf.kind = FILE_KIND_BIN;
}

word memo_find_first()
{/* Find first record of file */
  memo_reset();
  return LibLFileFindNext(&fb, &finf, 0x00);
}

word memo_find_next()
{/* Find next record of file (back-mode) */
  return LibLFileFindPrev(&fb, &finf, 0x00);
}


word memo_read(void)
{/* Read record from flash */
  if (!LibLFileRead(&fb, &finf)) 
    return 0;
  correct_settings();
  return 1;
}

word memo_load_next(void) 
{/* Load next record */

  if (memo_find_next()) 
    return memo_read();
  else
    return 0;
}

word memo_load_first(void) 
{/* Load first record */
  if (memo_find_first())
    return memo_read();
  else
    return 0;
}

word memo_save_curr(void) 
{/* Save current record */
  fb.fbuf.bin.char_num = g_stat->buf_size+1+sizeof(t_stat);
  return LibLFileWrite(&fb, &finf);
}

word memo_save_new(void) 
{/* Save new record */
  memo_reset();
  finf.fp = 0xFFFF;
  if (!memo_save_curr())
    return 0;
  new_file = 0;
  return 1;
}

word memo_save(void) 
{/* Save record */
  if (new_file) 
    return memo_save_new();
  else 
    return memo_save_curr();
}

word memo_delete(word num) 
{/* Delete record with specifed number */
  int i;

  memo_reset();

  if (!memo_find_first())
    return 0;
  for (i=0; i<num; i++) 
  {
    if (!memo_find_next()) 
      return 0;
  }
  LibLFileRemove(&fb, &finf);
  return 1;
}

/* Get action-control info (@see: leverpush.a86) */
int far getleverpush();


void wait_press(void) 
{/* Wait for screen pressing */
  do {
    LibTchWait(&g_tch);
  } while(!(g_tch.act & (ACT_MAKE|ACT_REPEAT)));
}


void wait_move(void) 
{/* Wait for stylus moving */
  LibTchWait(&g_tch);
}


void PollEvent(TCHSTS far* tsts, byte event_mask)
{/* Get stylus info */
  union REGS reg;
  reg.x.ax = 0x0200 | event_mask;
  reg.x.di = FP_OFF(tsts);
  reg.x.es = FP_SEG(tsts);
  int86(0x50,&reg,&reg);
}


int is_scroll(int not_lever)
{/* Detect scrolling */
  if (not_lever)
  {
    TCHSTS tch2;
    PollEvent(&tch2,1);
    return tch2.x!=-1&&tch2.y!=-1;
  }
  return getleverpush();
}

word pad_get_line(word line_num, byte far **line) 
{/* Get one line with specifed number from buffer */
  /* Lerm
  return - offset of line
  line   - pointer to line
  */
  word pos = 0;
  word start_line;
  word i;
  if (line_num<=level)
    start_line=line_num & 0xFFC0;
  else
    start_line=level;

  pos=points[start_line>>6];
  *line = &g_buf[pos];
  
  for (i=start_line; i<line_num;) 
  {
    byte b = g_buf[pos];
    if (b == BUFFER_END) 
    {
      *line = &g_buf[pos];
      return 0;
    }
    pos += (b & 0x1F);
    i++;
    if ((i & 0x3F)==0)
    {
      points[i>>6]=pos;
      level=i;
    }

  }
  *line = &g_buf[pos];
  return pos;
}

byte pad_get_next_line(byte far **line) 
{/* Move to next line in buffer */
  /* Lerm
  return - size of current line
  line - pointer to next line
  */
  byte len;
  if (**line == BUFFER_END) 
    return 0;
  len = **line & 0x1F;
  *line += len;
  return len;
}

/* Compress one line (@see: comp_poa.a86) */
byte far pad_compress_line(byte far *dest, byte far* src);
/*
 This function is correct. You can use it instead of asm-function. Lerm.
byte pad_compress_line(byte far *dest, byte far* src) 
{
  byte pos=1;
  byte len=0;
  byte i;
  byte curr=0;

  if (src[0] & 0x80) 
  {
    dest[pos++] = 0;
    curr=1;
  }
  for (i=0; i<20; i++) 
  {
    byte s = src[i];
    if (s == (curr ? 0xFF : 0)) 
    {
      len += 8;
    } 
    else 
    {
      byte j;
      for (j=0; j<8; j++) 
      {
        if ((s & 0x80) == (curr ? 0x0 : 0x80)) 
        {
          curr = 1-curr;
          if (pos == 20) 
            goto nocompress;
          dest[pos++] = len;
          len=0;
        } 
	s <<= 1;
        len++;
      }
    }
  }
  dest[0] = LINE_COMPRESSED | pos;
  return pos;

  nocompress:
  dest[0] = LINE_FREE | 21;
  far_memmove(&dest[1], src, 20);
  return 21;
}
*/

/* Uncompress one line (@see: comp_poa.a86) */
byte far pad_decompress_line(byte far *dest, byte far *src);
/*
 This code fragment is invalid. Don't uncomment it! Only for information purposes. Lerm.
byte pad_decompress_line(byte far *dest, byte far *src)
{
  byte s = src[0];
  if ((s & LINE_COMPRESSED) == LINE_COMPRESSED) 
  {
    byte len=s & 0x1F;
    byte i=2;
    byte pos=0;    
    far_memset(dest, 0, 20);
    if (len==1) 
      return 1;
    s=src[1];
    if (s) 
    {
      pos = s;
    } 
    else 
    {
      if (len==2) 
      {
	far_memset(dest, 0xFF, 20);
        return 2;
      }
    }
    s=src[i];
    do 
    {
      byte start_pos = pos;
      byte stop_pos = pos+s;
      byte start_byte = start_pos >> 3;
      byte stop_byte = stop_pos >> 3;
      byte start_bit = start_pos & 0x07;
      byte stop_bit = stop_pos & 0x07;
      if (start_byte == stop_byte) 
      {
	dest[start_byte] |= (0xFF >> (8-stop_bit+start_bit)) << (8-stop_bit);
      } 
      else 
      {
        byte j;
        byte curr_byte = start_byte;
	dest[curr_byte++] |= 0xFF >> start_bit;
	for (j=0; j<stop_byte-start_byte-1; j++) 
        {
	  dest[curr_byte++] = 0xFF;
	}
	dest[curr_byte++] |= 0xFF << (8-stop_bit);
      }  
      pos+=s;
      i++;
      if (i<len) 
      {
        pos += src[i++];
	if (i==len) 
        { 
          s = 160-pos; 
          i--;  
        }
        else 
          s = src[i];
      }
    } while(i<len);
    return len;
  } 
  else 
  {
    far_memmove(dest, &src[1], 20);
    return 21;
  }
}
*/

void pad_get_scr(void) 
{/* Load from buffer to screen-area */
  byte far *line;
  byte i=0;
  pad_get_line(g_stat->line, &line);
  memset(&g_scr[4], 0, 150*20);
  while(*line != 0xFF && i<150) 
  {
    line+=pad_decompress_line(&g_scr[4+i*20], line);
    i++;
  }
  if (i<150) 
  {
    byte t = 150-i;
    far_memset(line, 0xE1, t);
    line[t] = 0xFF;
    g_stat->lines += t;
    g_stat->buf_size += t;
    need_save_scroll=1;
  }
}

char pad_save_scr(void) 
{/* Save screen-area to buffer */
  byte buffer[150*21];
  word scr_size = 0;
  word old_size = 0;
  byte i;
  byte far *start_line;
  byte far *line;
  word pos = pad_get_line(g_stat->line, &line);
  start_line = line;
  for (i=0; i<150; i++) 
  {
    scr_size += pad_compress_line(&buffer[scr_size], &g_scr[4+i*20]);
    old_size += pad_get_next_line(&line);
  }
  if (g_stat->buf_size + scr_size - old_size > MAX_BUFFER_LEN) 
  {
    pad_get_scr();
    return 0;
  }
  if (*line == BUFFER_END) 
  {
    start_line[scr_size] = BUFFER_END;
    g_stat->lines = g_stat->line + 150;
  } 
  else 
  {
    far_memmove(&start_line[scr_size], line, g_stat->buf_size-pos-old_size+1);
  }
  far_memmove(start_line, buffer, scr_size);
  g_stat->buf_size -= old_size;
  g_stat->buf_size += scr_size;
  level=g_stat->line&0xFFC0;
  return 1;
}

void pad_draw(byte line) 
{/* Transfer screen-area to display */
  LibPutGraph(0, 0, g_scr);
  if (g_stat->line < 30 && line) 
  {
    byte y = 30 - g_stat->line;
    LibGdsLineMesh(0, y, 159, y);
  }
}

void pad_insert_lines(byte y, byte h) 
{/* Insert lines in buffer */
  byte far *line;
  word abs_y = g_stat->line + y;
  word pos = pad_get_line(abs_y, &line);
  if (g_stat->buf_size + h < MAX_BUFFER_LEN) 
  {
    far_memmove(line+h, line, g_stat->buf_size - pos + 1);
    far_memset(line, 0xE1, h);
    g_stat->buf_size += h;
    g_stat->lines += h;
  }
  level=g_stat->line&0xFFC0;
}

void pad_delete_lines(byte y, byte h) 
{/* Delete lines from buffer */
  if (y>=h)
  {
    byte far *line1;
    byte far *line2;
    word end_line = g_stat->line + y;
    word pos1 = pad_get_line(end_line-h, &line1);
    word pos2 = pad_get_line(end_line, &line2);

    far_memmove(line1, line2, g_stat->buf_size - pos2 + 1);
    g_stat->buf_size -= pos2-pos1;
    g_stat->lines -= h;
    level=g_stat->line&0xFFC0;
  }
}

void pad_init() 
{/* Init header for new record */
  /**
  int i=0;
  for(;i<18;i++)
    ((byte far *)g_stat)[i]=default_header[i];
  /*isVer2=1;*//* Reserved */
  /**/far_memmove(g_stat,default_header,18);/**/
  memset(g_scr, 0, 3004);
  g_scr[0] = 160;
  /*g_scr[1] = 0;*/
  g_scr[2] = 150;
  /*g_scr[3] = 0;*/
  g_buf[0] = 0xFF;
  points[0] = 0;
  level=0;

}

void file_draw_header(byte y) 
{/* Draw file caption */
  byte far *line;
  byte i=0;
  pad_get_line(0, &line);
  memset(&g_file_scr[4], 0, 29*20);
  g_file_scr[0] = 160;
  g_file_scr[1] = 0;
  g_file_scr[2] = 29;
  g_file_scr[3] = 0;

  while(*line != 0xFF && i<29) 
  {
    line+=pad_decompress_line(&g_file_scr[4+i*20], line);
    i++;
  }
  LibPutGraph(0, y, g_file_scr);
}

void dispString(char* str)
{/* Display string in window */
    int strSize=LibGetProStrSize(IB_PFONT1,str);
    LibPutFarData(15,54,107);
    LibPutProStr(IB_PFONT1,15+((128-strSize)>>1),75,str,strSize);
    LibPutDisp();
    set_touch(TchListFull);
    wait_press();
    set_touch(full_screen ? TchListMenu : TchListDraw);
}

byte* field_write(int number, byte* string, int field)
{/* Convert number to string (with filling) */
    /* Size optimizing (~200 bytes) (UNV)
    byte str[7];
    int i,j;
    itoa(number,str,10);
    i=strlen(str);
    if (i<field)
        for(j=field-i; j; j--)
            *(string++)=' ';
    for(j=0; j<i; j++)
        *(string++)=str[j];
    *string=0;
    return string;
    */
    byte str[7];
    int i, j;
    int n = number;
    char sign = 0;
    if (n < 0)
    {
        sign++;
        n = -n;
    }
    i = 0;
    do { str[i++] = n%10+'0'; } while ((n/=10) > 0);
    str[i] = '-';
    i += sign;

    for(j=field-i; j>0; j--)
        *(string++) = ' ';
    while (i>0)
        *(string++) = str[--i];
    *string = 0;
    return string;
}

void file_draw(word file, word sel, word num) 
{/* Draw records captions */
  word i;
  byte end_draw = memo_find_first();
  byte number[10];
  int  length;

  if (file>1)
    for (i=0; i<file-1; i++)
    {
      memo_find_next();
    }

  memo_read();

  LibPutGraphO(0,0,scroll_but_up,IB_GPOVER);
  LibPutGraphO(0,144,scroll_but_down,IB_GPOVER);

  if (file==0)
    LibPutGraphO(67,10,scroll_but_up2,IB_GPOVER);
  if (file+3>=num)
    LibPutGraphO(67,145,scroll_but_down2,IB_GPOVER);


  for (i=0; i<4&&file+i<=num; i++)
  {
    int y_coord=i*32+16;
    if (file != 0 || i!=0)
    {
      file_draw_header(y_coord+1);

      field_write(file+i,number,0);
      length=strlen(number);
      LibGdsClr(157-length*4,y_coord+2,157,y_coord+8);
      LibPut35Str(158-length*4,y_coord+3,number);

      if (file + i == sel)
         LibGdsReverse(0, y_coord+1, 159, y_coord+(i==3?29:30));
      if (!memo_load_next())
        end_draw=0;
    }
    else
    {
      LibPutGraphO(48,y_coord+6,new_icon,IB_GPOVER);
      if (sel == 0)
        LibGdsReverse(0, y_coord+1, 159, y_coord+30);
    }
    if (i<3)
      LibPutGraphO(0,y_coord+30,scroll_delim,IB_GPOVER);
    if (!end_draw)
      break;
  }
  LibGdsLine(0,16,0,143);
  LibGdsLine(159,16,159,143);
}

word file_num(void) 
{/* Get records count */
  /* Hmm... It's smaller... */
  word i = 1;
  if (!memo_find_first()) 
    return 0;
  while (memo_find_next()) 
    i++;
  return i;
  /**/
  /* Comments reason: size optimization *
  return LibLGetFileCnt(&fb);
  /**/
}

void file_open(word file) 
{/* Load file */
  word i;
  if (!memo_find_first()) 
  {
    dispString("Cannot find first!");
    return;
  }
  for (i=0; i<file; i++) 
  {
    if (!memo_find_next()) 
    {
      byte str[20]="Found only ";    
      field_write(i,str+12,0);
      dispString(str);
      return;
    }
  }
  memo_read();
  g_scr[0] = 160;
  g_scr[1] = 0;
  g_scr[2] = 150;
  g_scr[3] = 0;
}

void pad_line(int x1, int y1, int x2, int y2);

void file_scroll(int direction, int not_lever, int num)
{/* Scroll in main menu */
  int e=1;
  int i=0;
  if (!not_lever)
  {
    if (direction&&sel==0)
    { 
      sel=num+1;
    }
    if (direction==0&&sel==num)
    {
      sel=-1;
    }
  }
  do  
  {
    if (not_lever)
    {
      if (direction)
      {
        if (file)
          file--;
        else
          e=0;
      }
      else
      {
        if (file+3<num)
          file++;
        else
          e=0;
      }
    }
    else
    {
      if (direction)
      {
        if (sel) 
          sel--;
        else
          e=0;
        if (sel < file) 
          file=sel;
        else
          if (sel>file+3)
            file=sel-3;
      }
      else
      {
        if (sel < num) 
          sel++;
        else
          e=0;
        if (sel < file) 
          file=sel;
        else
          if (sel>file+3)
            file=sel-3;
      }
    }
    if (e)
    {
      LibClrDisp();
      file_draw(file, sel, num);
      if (not_lever)
        if (direction)
          LibGdsReverse(0, 0, 159, 15);
        else
          LibGdsReverse(0, 144, 159, 159);
      LibPutDisp();
      if (file<16)
      {
        if (i<3)
          LibWait(IB_250MWAIT);        
        else
          LibWait(IB_125MWAIT);
        i++;
      }       
    } 
  } while (is_scroll(not_lever)&&e);
  while(is_scroll(not_lever))
  {}
}

byte file_select(void) 
{/* Main menu */
  word num = file_num();
  word newsel = 0;
  word line = 0;
  dword act=0;
  byte flag=0;
  int start_x,start_y,x2,y2;
  
  new_file = 0;

  /* If no record then start in painting mode */
  /*
  if (num == 0) 
  {
    new_file = 1;
    return 1;
  }
  */

  if (sel>num)
  {
    sel=0;
    file=0;
  }

  do {
    LibClrDisp();
    file_draw(file, sel, num);
    g_stat->pen_size = 1;
    g_stat->pen_mesh = 0;
    g_stat->eraser_size = 1;
    LibPutDisp();
    wait_press();

    switch(g_tch.obj)
    {
      case OBJ_LPSW_UP:
        file_scroll(1,0,num);
        break;
      case OBJ_IC_MENU_UP:        
        file_scroll(1,1,num);
        break;
      case OBJ_LPSW_DOWN:
        file_scroll(0,0,num);
        break;
      case OBJ_IC_MENU_DOWN:
        file_scroll(0,1,num);
        break;
      case OBJ_LPSW_LEFT:
        sel=0;
        file=0;
        break;
      case OBJ_LPSW_RIGHT:
        sel=num;
        if (num>3)
          file=num-3;
        break;
      case OBJ_IC_MENU_L1: case OBJ_IC_MENU_L2:
      case OBJ_IC_MENU_L3: case OBJ_IC_MENU_L4:
        line=g_tch.obj-OBJ_IC_MENU_L1;
        start_x = g_tch.x;
        start_y = g_tch.y;
        newsel = file+line;
        if (newsel<=num)
        {
          if (newsel!=sel)
          {
            g_stat->pen_type = PEN_DRAW;
            flag=0;
          }
          else
          {
            g_stat->pen_type = PEN_ERASE;
            flag=1;
          }
          
          do {
    	    x2 = g_tch.x;
 	    y2 = g_tch.y;
  	    wait_move();
	    act=g_tch.act;
	    
	    if (act & (ACT_MOVE_OUT|ACT_MOVE_IN)&&flag)
	      g_stat->pen_type^=1;
	    
	    pad_line(g_tch.x, g_tch.y, x2, y2);
	    /*LibGdsLineCmp(g_tch.x, g_tch.y, x2, y2);*/
	    LibPutDisp();
          } while(!(act & (ACT_BREAK_IN|ACT_BREAK)));

          if (act==ACT_BREAK_IN)
          {
            if (newsel)
            {
              if (abs(g_tch.x-start_x)>80)
              {
  	        memo_delete(newsel - 1);
  	        num=file_num();
  	        if (file+3>num&&file)
  	        {
  	          file--;
  	        }
                if (sel>num)
                  sel--;
	        break;
              }
              if (sel==newsel)
              {
                file_open(newsel-1);
                return 1;
              }
            }
            else
            {
              if(sel==0)
              {
                pad_init();
                new_file = 1;
                sel=1;
                file=0;
                return 1;
              }
            }
          }
          sel=newsel;
        }
        break;
      case OBJ_LPSW_PUSH:
        if (sel == 0) 
        {
	  pad_init();
	  new_file = 1;
        } 
        else 
        {
          file_open(sel-1);
        }
        return 1;
      case OBJ_HIC_ESC:
        return 0;
    }
  } while(1);
}

void menu_draw_stat(int y) 
{/* Draw statusbar */
  byte str[40];
  byte *str_temp=str;
 
  str_temp=field_write(g_stat->line,str_temp,5);
  *(str_temp++)='/';
  str_temp=field_write(g_stat->lines,str_temp,5);
  *(str_temp++)=' ';
  field_write(g_stat->buf_size,str_temp,5);
  LibGdsClr(91,y,159,y+5);
  LibPut35Str(91, y, str);
}

void menu_draw(void) 
{/* Draw toolbar */
  byte draw_style=g_stat->draw_style;
  LibGdsClr(0, 150, 159, 159);
  LibPutGraph(  1,151,icon_pen[g_stat->pen_mesh&&1][g_stat->pen_size]);
  LibPutGraph( 16,151,icon_eraser[g_stat->eraser_size]);
  LibPutGraph( 31,151,icon_correct);
  LibPutGraph( 61,151,icon_mesh[g_stat->pen_mesh]);
  LibPutGraph( 76,151,icon_undo);

  switch (g_stat->pen_type) 
  {
    case PEN_DRAW:    
      LibPutGraph( 46,151,icon_style[draw_style]);
      LibGdsReverse( 0, 151, 13, 159); 
      break;
    case PEN_ERASE: 
      LibPutGraph( 46,151,icon_style[g_stat->erase_style]);
      LibGdsReverse(15, 151, 28, 159); 
      break;
    default:          
      LibPutGraph( 46,151,icon_style[draw_style]);
      LibGdsReverse(30, 151, 43, 159);
  }
  menu_draw_stat(154);
}

void pad_up(int scroll_step, int ext) 
{/* Scroll image up */
  byte empty = 1;
  byte far *start_line;
  byte far *line;
  word pos;

  if (g_stat->line > scroll_step) 
    g_stat->line -= scroll_step;
  else 
    g_stat->line = 0;
  pad_get_scr();

  pos = pad_get_line( g_stat->line + 150, &line);
  start_line = line;
  while(*line != BUFFER_END) 
  {
    if (*line != 0xE1) 
    {
      empty = 0;
      break;
    }
    pad_get_next_line(&line);
  }
  if (empty && *start_line != BUFFER_END) 
  {
    *start_line = BUFFER_END;
    g_stat->lines = g_stat->line + 150;
    g_stat->buf_size = pos;
    need_save_scroll=1;
    /*level=g_stat->line&0xFF80;*/
  }

  pad_draw(1);
  if (ext)
  {
    menu_draw_stat(144);
    LibPutGraphO(90,151,d_up,IB_GPREV);
    LibPutGraphO(125,151,d_down,IB_GPOVER);
  }
  else
  {
    menu_draw_stat(154);
  }
  LibPutDisp();
}

void pad_down(int scroll_step, int ext) 
{/* Scroll image down */
  if (g_stat->buf_size >= MAX_BUFFER_LEN) 
    return;
  g_stat->line += scroll_step;
  pad_get_scr();
  pad_draw(1);
  if (ext)
  {
    menu_draw_stat(144);
    LibPutGraphO(90,151,d_up,IB_GPOVER);
    LibPutGraphO(125,151,d_down,IB_GPREV);
  }
  else
  {
    menu_draw_stat(154);
  }
  LibPutDisp();
}

void pad_pixel(int x, int y) 
{/* Draw pixel */
  byte mesh  = g_stat->pen_mesh;
  byte draw  = (g_stat->pen_type == PEN_DRAW);
  byte size  = draw ? g_stat->pen_size : g_stat->eraser_size;
  byte pen   = size;
  byte tsize = (size == 2) ? 4 : size+1;
  int mode;

  if (x<0)
  {
    x = 0;
  }
  else if (x>160-tsize) 
  {
    x = 160-tsize;
  }

  if (y<0) 
  {
    y = 0;
  }
  else if (full_screen)
  {
    if (y>160-tsize) 
      y = 160-tsize;
  }
  else
  {
    if (y>150-tsize) 
      y = 150-tsize;
  }
  
  if (draw) 
  {
    if (mesh) 
    {
      pen += 3;
      if ((x + y + g_stat->line) % 2 != 0)
        pen += 3;
    }
    mode = IB_GPOR;
  } 
  else 
  {
    mode = IB_GPAND;
  }

  if (mesh==1 && draw)
  {
    LibPutGraphO(x, y, g_mask[size], IB_GPAND);
  }

  LibPutGraphO(x, y, g_pen[pen], mode);

}

void pad_line(int x1, int y1, int x2, int y2) 
{/* Draw line */
  int x=x1, y=y1;
  int dx=x2-x1, dy=y2-y1;
  int incx=1, incy=1;
  int balance;

  if (dx<0)
  {
    dx = -dx;
    incx = -incx;
  }
  if (dy<0)
  {
    dy = -dy;
    incy = -incy;
  }

  if (dx >= dy)
  {
    balance = dx>>1;

    while (x != x2)
    {
      pad_pixel(x, y);
      if ((balance -= dy) < 0)
      {
        y += incy;
        balance += dx;
      }
      x += incx;
    } 
    pad_pixel(x, y);
  }
  else
  {
    balance = dy>>1;

    while (y != y2)
    {
      pad_pixel(x, y);
      if ((balance -= dx) < 0)
      {
        x += incx;
        balance += dy;
      }
      y += incy;
    } 
    pad_pixel(x, y);
  }
}

void pad_box(int x1, int y1, int x2, int y2)
{/* Draw rectangle */
  pad_line(x1, y1, x1, y2);
  pad_line(x2, y1, x2, y2);
  pad_line(x1, y1, x2, y1);
  pad_line(x1, y2, x2, y2);
}

void pad_straight_line(int x1, int y1, int x2, int y2)
{/* Draw straight line */
 /* TODO: start position correcting */
  int dx=x2-x1, dy=y2-y1;
  int abs_dx=(dx>0)?dx:-dx;
  int abs_dy=(dy>0)?dy:-dy;

  if (abs_dx > abs_dy*2)
  {
    pad_line(x1, y1, x2, y1);
  }
  else if (abs_dy > abs_dx*2)
  {
    pad_line(x1, y1, x1, y2);
  }
  else
  {
    if (abs_dx <= abs_dy)
    {
      pad_line(x1, y1, x2, (dy>0)?(y1+abs_dx):(y1-abs_dx));
    }
    else
    {
      pad_line(x1, y1, (dx>0)?(x1+abs_dy):(x1-abs_dy), y2);
    }
  }
}

void pad_ellipse(int x1, int y1, int x2, int y2)
{/* Draw ellipse */
  int dx=x2-x1;
  int dy=y2-y1;
  int a=abs(dx), b=abs(dy);
  int xa=x1+(dx>>1), xb=x1+((dx+1)>>1);
  int ya=(dy>0)?y1:y2, yb=(dy>0)?y2:y1;
  long as=a*a, bs=b*b;
  long d=0, x2bspbs=bs, y2asmas=b*as-as;
  long xinc = 2*bs, yinc = 2*as;

  if (as<4 || bs<4)
    return;

  do
  {
    pad_pixel(xa, ya);
    pad_pixel(xa, yb);
    pad_pixel(xb, ya);
    pad_pixel(xb, yb);

    d += x2bspbs;
    xa--;
    xb++;
    x2bspbs += xinc;
    if (2*d-y2asmas>0)
    {
      d -= y2asmas;
      ya++;
      yb--;
      y2asmas -= yinc;
    }
  } while (y2asmas+2*as > x2bspbs-bs);
  while (ya<=yb)
  {
    pad_pixel(xa, ya);
    pad_pixel(xa, yb);
    pad_pixel(xb, ya);
    pad_pixel(xb, yb);

    d -= y2asmas;
    ya++;
    yb--;
    y2asmas -= yinc;
    if (2*d+x2bspbs<=0)
    {
      d += x2bspbs;
      xa--;
      xb++;
      x2bspbs += xinc;
    }
  }
}

void scroll(int direction, int not_lever)
{/* Scroll image */
  int scroll_step=2;
  pad_save_scr();
  while (is_scroll(not_lever)) 
  {
    if (scroll_step < 160) 
      scroll_step++;

    if (direction)
      pad_up(scroll_step>>1, not_lever);
    else
      pad_down(scroll_step>>1, not_lever);
  }
  pad_draw(1);
}

void reverseIcon(int psel)
{/* Reverse icon (exit confirmation panel) */
  int n=psel*32;
  if (psel==-1)
    return;
  LibGdsReverse(32+n,62,59+n,85);
  LibGdsReverse(60+n,62,60+n,62);
  LibGdsReverse(32+n,86,32+n,86);
}

int confirm_exit()
{/* Confirm exit from drawing mode */
  int psel=-1,e=1,csel;
  dialog_mode=1;
  pad_save_scr();
  pad_draw(1);
  menu_draw();
  LibPutGraphO(30, 60, panel, IB_GPOVER);
  LibPutDisp();
  set_touch(TchListPanel);
  do{
    wait_move();
    switch (g_tch.obj)
    {
      case OBJ_IC_PSAVE:
      case OBJ_IC_PCLOSE:
      case OBJ_IC_PEDIT:
        csel=g_tch.obj-OBJ_IC_PSAVE;
        if ((g_tch.act & (ACT_BREAK_IN | ACT_MOVE_OUT))&& psel==csel)
          e=0;
        else
          if (g_tch.act==ACT_BREAK)
          {
            reverseIcon(psel);
            psel=-1;
            LibPutDisp();
          }
          else
          {
            reverseIcon(psel);
            psel=csel;
            reverseIcon(psel);
            LibPutDisp();
          }
        break;
      case OBJ_IC_SCREEN:
      case OBJ_HIC_ESC:
        psel=2;
        e=0;
    }
  } while (e);
  set_touch(TchListDraw);
  pad_draw(1);
  LibPutDisp();
  dialog_mode=0;
  return psel;
}

char pad_check(void) 
{/* Drawing mode */
  int cur_x = g_tch.x;
  int cur_y = g_tch.y;
  int start_x = cur_x;
  int start_y = cur_y;
  int old_x = cur_x;
  int old_y = cur_y;
  byte pen_type=g_stat->pen_type;
  byte pen_size=g_stat->pen_size;
  byte eraser_size=g_stat->eraser_size;
  byte pen_mesh=g_stat->pen_mesh;
  byte draw_style=pen_type==PEN_ERASE?g_stat->erase_style:g_stat->draw_style;
  byte size = pad_draw ? pen_size : eraser_size;
  byte tsize = size == 2 ? 4 : size+1;

  switch (g_tch.obj)
  {
    case OBJ_IC_PAD:
      need_save=1;
      pad_save_scr();
      pad_draw(0);
      do {
        cur_x=g_tch.x;
        cur_y=g_tch.y;

        if (cur_x>(160-tsize))
          cur_x=160-tsize;
        else
          if (cur_x<0)
            cur_x=0;

        if (cur_y>(150-tsize))
          cur_y=150-tsize;
        else
          if (cur_y<0)
            cur_y=0;

        if (pen_type == PEN_CORRECT) 
        {
	  if (cur_y > old_y)
	    pad_insert_lines(old_y, cur_y - old_y);
	  if (cur_y < old_y)
	    pad_delete_lines(old_y, old_y - cur_y);
	  pad_get_scr();
	} 
        else 
          if (draw_style == STYLE_CURVE)
          {
            pad_line(cur_x, cur_y, old_x, old_y);
          }
          else
          {
            pad_draw(0);
            switch (draw_style)
            {
              case STYLE_LINE:
                pad_line(start_x, start_y, cur_x, cur_y);
                break;
              case STYLE_STRAIGHT:
                pad_straight_line(start_x,start_y,cur_x,cur_y);
                break;
              case STYLE_BOX:
                pad_box(start_x,start_y,cur_x,cur_y);
                break;
              case STYLE_ELLIPSE:
                pad_ellipse(start_x, start_y, cur_x, cur_y);
            }
          }
	old_x = cur_x;
	old_y = cur_y;
	if (pen_type == PEN_CORRECT) 
        {
	  pad_draw(1);
	  LibGdsLine(0, cur_y, 159, cur_y);
	}
	LibPutDisp();
        wait_move();
      } while(!(g_tch.act & (ACT_BREAK | ACT_BREAK_IN)));

      if (pen_type != PEN_CORRECT) 
      {
        LibGetGraph(0, 0, 160, 150, &g_scr[4]);
      }
      pad_draw(1);
      break;
    case OBJ_IC_MENU:
      if (cur_x < 15)
      {/* Pen */
        if (pen_type == PEN_DRAW) 
        {
	  if (++pen_size > 2) 
            pen_size = 0;
        }
        else
        {
          pen_type = PEN_DRAW;
        }
      } 
      else if (cur_x < 30) 
      {/* Eraser */
        if (pen_type == PEN_ERASE ) 
        {
          if (++eraser_size > 2) 
            eraser_size = 0;
        }
        else
        {
          pen_type = PEN_ERASE;
        }
      } 
      else if (cur_x < 45) 
      {/* Correct line */
        pen_type = PEN_CORRECT;
      } 
      else if (cur_x < 60)
      {/* Drawing style */
        if (++draw_style>STYLE_ELLIPSE)
          draw_style=STYLE_CURVE;
      }
      else if (cur_x < 75) 
      {/* Line type - solid/mesh/mesh2 */
        if (++pen_mesh>2)
          pen_mesh=0;
      } 
      else if (cur_x < 90) 
      {/* Undo */
        if (need_save)
        {
          pad_get_scr();
          pad_draw(1);
        }
      }
      else if (cur_x < 125)
      {/* Scroll up */
        scroll(1,1);
      }
      else
      {/* Scroll down */
        scroll(0,1);
      }
      break;
    case OBJ_LPSW_UP:
      scroll(1,0);
      break;
    case OBJ_LPSW_DOWN:
      scroll(0,0);
      break;
    case OBJ_LPSW_LEFT:
      if (pen_type == PEN_DRAW) 
      {
        if (++pen_size > 2) 
          pen_size = 0;
      }
      else if (pen_type == PEN_ERASE) 
      {
	if (++eraser_size > 2) 
          eraser_size = 0;
      }
      break;
    case OBJ_LPSW_RIGHT:
      if (++draw_style > STYLE_ELLIPSE)
        draw_style=STYLE_CURVE;
      break;
    case OBJ_LPSW_PUSH:
      pen_type = (pen_type == PEN_DRAW) ? PEN_ERASE : PEN_DRAW;
      break;
    case OBJ_HIC_MBAR:
      if (need_save || new_file || need_save_scroll)
      {
        pad_save_scr();
        memo_save();
      }
      return 0;
    case OBJ_HIC_ESC:
      if (need_save /*|| new_file*/)
      {
        int res=confirm_exit();
        if (res==2)
          return 1;
        if (res==0)
          memo_save();
      }
      return 0;
    default:
      return 1;
  }
  if (g_stat->pen_type==pen_type)
  {
    if (pen_type==PEN_ERASE)
      g_stat->erase_style=draw_style;
    else
      g_stat->draw_style=draw_style;
  }
  else
    g_stat->pen_type=pen_type;

  g_stat->pen_size=pen_size;
  g_stat->eraser_size=eraser_size;
  g_stat->pen_mesh=pen_mesh;
  
  /*need_save=1;*/
  menu_draw();
  LibPutDisp();
  return 1;
}

int far mode_hook(unsigned int iRegAX, unsigned int iRegBX)
{/* Mode hook handler */
  switch (iRegAX >> 8)
  {
      case IB_MDWINON:
	if (call_shot == 0) 
        {
	  byte far *ptr;
	  word far *p;          
	  call_shot = 1;
	  g_shot.x=0;
	  g_shot.y=0;
	  g_shot.x_size=156;
	  g_shot.y_size=156-29;
	  g_shot.p_x=2;
	  g_shot.p_y=29;
	  pad_draw(0);
	  LibScrShot(&g_shot);

/*	  ptr = LibDualWin(0x0900, 4, &g_shot);
	  LibPutMsgDlg("%x:%x", FP_SEG(ptr), FP_OFF(ptr)); */
	  p = MK_FP(0xc400, 0x07c4);
	  if ((p[0] == 0x609c) && (p[1] == 0x1e06)) 
          {
	    /* PV-Sx60 */
	    ptr = MK_FP(0x1240, 79*20+16);
	  } 
          else 
          {
	    /* Others */
	    ptr = MK_FP(0x1240, 81*20+8);
	  }

	  far_memmove(&g_scr[4], ptr, (156-29)*20);
	  pad_draw(1);
	  LibPutDisp();
	  call_shot = 0;
	  need_save = 1;
	  return 1;
	}
	return 0;
      case IB_OPENSYNCJP: return 0;
      case IB_MJUMP:
	if ((iRegBX & 0xFF00) == IW_MCOM) 
        {
	  return 0;
	}
      default:
        if (dialog_mode)
        {
          pad_draw(1);
          menu_draw();
          LibPutGraphO(30, 60, panel, IB_GPOVER);
          LibPutDisp();
          return 1;
        }
        if (need_save)
        {          
          int res=confirm_exit();
          if (res==2)
            return 1;
          if (res==0)
            memo_save();
          /* Lerm: remove...
          need_save=0;
          */
	}
	return 0;
  }

  return 0;
}

void main(void) 
{
  byte work=1;
 
  setmodehook(&mode_hook);

  if (!memo_init()) 
  {
    dispString("Can't create IDraw file!");
  }
  else
  {
    do {
      pad_init();
      full_screen=1;
      set_touch(TchListMenu);
      work=file_select();
      if (work)
      {
        pad_get_scr();
        pad_draw(1);
        menu_draw();
        LibPutDisp();
        call_shot = 0;
        full_screen = 0;
        set_touch(TchListDraw);
        do {
          wait_press();
        } while(pad_check());
        need_save = 0;
        need_save_scroll = 0;
        call_shot = 1;
      }
    } while(work);
  }

  releasemodehook();
  LibJumpMenu();
}